@@ -6,6 +6,7 @@
 //
 // SPDX-License-Identifier: MPL-2.0
 
+use futures::TryFutureExt;
 use gst::glib;
 use gst::prelude::*;
 use gst::subclass::prelude::*;
@@ -28,13 +29,21 @@ use std::sync::Mutex;
 use std::time::Duration;
 
 use crate::s3url::*;
-use crate::s3utils::{self, WaitError};
+use crate::s3utils::{self, duration_from_millis, duration_to_millis, RetriableError, WaitError};
 
 use super::OnError;
 
 const DEFAULT_MULTIPART_UPLOAD_ON_ERROR: OnError = OnError::DoNothing;
+// General setting for create / abort requests
+const DEFAULT_REQUEST_TIMEOUT_MSEC: u64 = 10_000;
+const DEFAULT_RETRY_DURATION_MSEC: u64 = 60_000;
+// This needs to be independently configurable, as the part size can be upto 5GB
 const DEFAULT_UPLOAD_PART_REQUEST_TIMEOUT_MSEC: u64 = 10_000;
 const DEFAULT_UPLOAD_PART_RETRY_DURATION_MSEC: u64 = 60_000;
+// CompletedMultipartUpload can take minutes to complete, so we need a longer value here
+// https://docs.aws.amazon.com/AmazonS3/latest/API/API_CompleteMultipartUpload.html
+const DEFAULT_COMPLETE_REQUEST_TIMEOUT_MSEC: u64 = 600_000; // 10 minutes
+const DEFAULT_COMPLETE_RETRY_DURATION_MSEC: u64 = 3_600_000; // 60 minutes
 
 struct Started {
     client: S3Client,
@@ -97,8 +106,12 @@ struct Settings {
     secret_access_key: Option<String>,
     metadata: Option<gst::Structure>,
     multipart_upload_on_error: OnError,
+    request_timeout: Option<Duration>,
+    retry_duration: Option<Duration>,
     upload_part_request_timeout: Option<Duration>,
     upload_part_retry_duration: Option<Duration>,
+    complete_upload_request_timeout: Option<Duration>,
+    complete_upload_retry_duration: Option<Duration>,
 }
 
 impl Settings {
@@ -167,12 +180,20 @@ impl Default for Settings {
             secret_access_key: None,
             metadata: None,
             multipart_upload_on_error: DEFAULT_MULTIPART_UPLOAD_ON_ERROR,
+            request_timeout: Some(Duration::from_millis(DEFAULT_REQUEST_TIMEOUT_MSEC)),
+            retry_duration: Some(Duration::from_millis(DEFAULT_RETRY_DURATION_MSEC)),
             upload_part_request_timeout: Some(Duration::from_millis(
                 DEFAULT_UPLOAD_PART_REQUEST_TIMEOUT_MSEC,
             )),
             upload_part_retry_duration: Some(Duration::from_millis(
                 DEFAULT_UPLOAD_PART_RETRY_DURATION_MSEC,
             )),
+            complete_upload_request_timeout: Some(Duration::from_millis(
+                DEFAULT_COMPLETE_REQUEST_TIMEOUT_MSEC,
+            )),
+            complete_upload_retry_duration: Some(Duration::from_millis(
+                DEFAULT_COMPLETE_RETRY_DURATION_MSEC,
+            )),
         }
     }
 }
@@ -216,8 +237,11 @@ impl S3Sink {
         let upload_id = &state.upload_id;
         let client = &state.client;
 
-        let upload_part_req_future =
-            || client.upload_part(self.create_upload_part_request(&body, part_number, upload_id));
+        let upload_part_req_future = || {
+            client
+                .upload_part(self.create_upload_part_request(&body, part_number, upload_id))
+                .map_err(RetriableError::Rusoto)
+        };
 
         let output = s3utils::wait_retry(
             &self.canceller,
@@ -278,7 +302,7 @@ impl S3Sink {
                 }
                 Some(gst::error_msg!(
                     gst::ResourceError::OpenWrite,
-                    ["Failed to upload part: {}", err]
+                    ["Failed to upload part: {:?}", err]
                 ))
             }
             WaitError::Cancelled => None,
@@ -313,16 +337,9 @@ impl S3Sink {
 
     fn create_complete_multipart_upload_request(
         &self,
-        started_state: &mut Started,
+        started_state: &Started,
+        completed_upload: CompletedMultipartUpload,
     ) -> CompleteMultipartUploadRequest {
-        started_state
-            .completed_parts
-            .sort_by(|a, b| a.part_number.cmp(&b.part_number));
-
-        let completed_upload = CompletedMultipartUpload {
-            parts: Some(std::mem::take(&mut started_state.completed_parts)),
-        };
-
         let url = self.url.lock().unwrap();
         CompleteMultipartUploadRequest {
             bucket: url.as_ref().unwrap().bucket.to_owned(),
@@ -369,45 +386,74 @@ impl S3Sink {
             Some(ref url) => url.clone(),
             None => unreachable!("Element should be started"),
         };
-        let abort_req = self.create_abort_multipart_upload_request(&s3url, started_state);
-        let abort_req_future = started_state.client.abort_multipart_upload(abort_req);
+        let abort_req_future = || {
+            let abort_req = self.create_abort_multipart_upload_request(&s3url, started_state);
+            started_state
+                .client
+                .abort_multipart_upload(abort_req)
+                .map_err(RetriableError::Rusoto)
+        };
 
-        s3utils::wait(&self.abort_multipart_canceller, abort_req_future)
-            .map(|_| ())
-            .map_err(|err| match err {
-                WaitError::FutureError(err) => {
-                    gst::error_msg!(
-                        gst::ResourceError::Write,
-                        ["Failed to abort multipart upload: {}.", err.to_string()]
-                    )
-                }
-                WaitError::Cancelled => {
-                    gst::error_msg!(
-                        gst::ResourceError::Write,
-                        ["Abort multipart upload request interrupted."]
-                    )
-                }
-            })
+        s3utils::wait_retry(
+            &self.abort_multipart_canceller,
+            Some(Duration::from_millis(DEFAULT_REQUEST_TIMEOUT_MSEC)),
+            Some(Duration::from_millis(DEFAULT_RETRY_DURATION_MSEC)),
+            abort_req_future,
+        )
+        .map(|_| ())
+        .map_err(|err| match err {
+            WaitError::FutureError(err) => {
+                gst::error_msg!(
+                    gst::ResourceError::Write,
+                    ["Failed to abort multipart upload: {:?}.", err]
+                )
+            }
+            WaitError::Cancelled => {
+                gst::error_msg!(
+                    gst::ResourceError::Write,
+                    ["Abort multipart upload request interrupted."]
+                )
+            }
+        })
     }
 
     fn complete_multipart_upload_request(
         &self,
         started_state: &mut Started,
     ) -> Result<(), gst::ErrorMessage> {
-        let complete_req = self.create_complete_multipart_upload_request(started_state);
-        let complete_req_future = started_state.client.complete_multipart_upload(complete_req);
+        started_state
+            .completed_parts
+            .sort_by(|a, b| a.part_number.cmp(&b.part_number));
 
-        s3utils::wait(&self.canceller, complete_req_future)
-            .map(|_| ())
-            .map_err(|err| match err {
-                WaitError::FutureError(err) => gst::error_msg!(
-                    gst::ResourceError::Write,
-                    ["Failed to complete multipart upload: {}.", err.to_string()]
-                ),
-                WaitError::Cancelled => {
-                    gst::error_msg!(gst::LibraryError::Failed, ["Interrupted during stop"])
-                }
-            })
+        let completed_upload = CompletedMultipartUpload {
+            parts: Some(std::mem::take(&mut started_state.completed_parts)),
+        };
+
+        let complete_req_future = || {
+            let complete_req = self
+                .create_complete_multipart_upload_request(started_state, completed_upload.clone());
+            started_state
+                .client
+                .complete_multipart_upload(complete_req)
+                .map_err(RetriableError::Rusoto)
+        };
+
+        s3utils::wait_retry(
+            &self.canceller,
+            Some(Duration::from_millis(DEFAULT_COMPLETE_REQUEST_TIMEOUT_MSEC)),
+            Some(Duration::from_millis(DEFAULT_COMPLETE_RETRY_DURATION_MSEC)),
+            complete_req_future,
+        )
+        .map(|_| ())
+        .map_err(|err| match err {
+            WaitError::FutureError(err) => gst::error_msg!(
+                gst::ResourceError::Write,
+                ["Failed to complete multipart upload: {:?}.", err]
+            ),
+            WaitError::Cancelled => {
+                gst::error_msg!(gst::LibraryError::Failed, ["Interrupted during stop"])
+            }
+        })
     }
 
     fn finalize_upload(&self, element: &super::S3Sink) -> Result<(), gst::ErrorMessage> {
@@ -463,20 +509,29 @@ impl S3Sink {
             _ => S3Client::new(s3url.region.clone()),
         };
 
-        let create_multipart_req = self.create_create_multipart_upload_request(&s3url, &settings);
-        let create_multipart_req_future = client.create_multipart_upload(create_multipart_req);
+        let create_multipart_req_future = || {
+            let create_multipart_req =
+                self.create_create_multipart_upload_request(&s3url, &settings);
+            client
+                .create_multipart_upload(create_multipart_req)
+                .map_err(RetriableError::Rusoto)
+        };
 
-        let response = s3utils::wait(&self.canceller, create_multipart_req_future).map_err(
-            |err| match err {
-                WaitError::FutureError(err) => gst::error_msg!(
-                    gst::ResourceError::OpenWrite,
-                    ["Failed to create multipart upload: {}", err]
-                ),
-                WaitError::Cancelled => {
-                    gst::error_msg!(gst::LibraryError::Failed, ["Interrupted during start"])
-                }
-            },
-        )?;
+        let response = s3utils::wait_retry(
+            &self.canceller,
+            Some(Duration::from_millis(DEFAULT_REQUEST_TIMEOUT_MSEC)),
+            Some(Duration::from_millis(DEFAULT_RETRY_DURATION_MSEC)),
+            create_multipart_req_future,
+        )
+        .map_err(|err| match err {
+            WaitError::FutureError(err) => gst::error_msg!(
+                gst::ResourceError::OpenWrite,
+                ["Failed to create multipart upload: {:?}", err]
+            ),
+            WaitError::Cancelled => {
+                gst::error_msg!(gst::LibraryError::Failed, ["Interrupted during start"])
+            }
+        })?;
 
         let upload_id = response.upload_id.ok_or_else(|| {
             gst::error_msg!(
@@ -656,6 +711,24 @@ impl ObjectImpl for S3Sink {
                     DEFAULT_MULTIPART_UPLOAD_ON_ERROR as i32,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
                 ),
+                glib::ParamSpecInt64::new(
+                    "request-timeout",
+                    "Request timeout",
+                    "Timeout for general S3 requests (in ms, set to -1 for infinity)",
+                    -1,
+                    std::i64::MAX,
+                    DEFAULT_REQUEST_TIMEOUT_MSEC as i64,
+                    glib::ParamFlags::READWRITE,
+                ),
+                glib::ParamSpecInt64::new(
+                    "retry-duration",
+                    "Retry duration",
+                    "How long we should retry general S3 requests before giving up (in ms, set to -1 for infinity)",
+                    -1,
+                    std::i64::MAX,
+                    DEFAULT_RETRY_DURATION_MSEC as i64,
+                    glib::ParamFlags::READWRITE,
+                ),
                 glib::ParamSpecInt64::new(
                     "upload-part-request-timeout",
                     "Upload part request timeout",
@@ -674,6 +747,24 @@ impl ObjectImpl for S3Sink {
                     DEFAULT_UPLOAD_PART_RETRY_DURATION_MSEC as i64,
                     glib::ParamFlags::READWRITE,
                 ),
+                glib::ParamSpecInt64::new(
+                    "complete-upload-request-timeout",
+                    "Complete upload request timeout",
+                    "Timeout for the complete multipart upload request (in ms, set to -1 for infinity)",
+                    -1,
+                    std::i64::MAX,
+                    DEFAULT_COMPLETE_REQUEST_TIMEOUT_MSEC as i64,
+                    glib::ParamFlags::READWRITE,
+                ),
+                glib::ParamSpecInt64::new(
+                    "complete-upload-retry-duration",
+                    "Complete upload retry duration",
+                    "How long we should retry complete multipart upload requests before giving up (in ms, set to -1 for infinity)",
+                    -1,
+                    std::i64::MAX,
+                    DEFAULT_COMPLETE_RETRY_DURATION_MSEC as i64,
+                    glib::ParamFlags::READWRITE,
+                ),
             ]
         });
 
@@ -750,19 +841,29 @@ impl ObjectImpl for S3Sink {
                 settings.multipart_upload_on_error =
                     value.get::<OnError>().expect("type checked upstream");
             }
+            "request-timeout" => {
+                settings.request_timeout =
+                    duration_from_millis(value.get::<i64>().expect("type checked upstream"));
+            }
+            "retry-duration" => {
+                settings.retry_duration =
+                    duration_from_millis(value.get::<i64>().expect("type checked upstream"));
+            }
             "upload-part-request-timeout" => {
                 settings.upload_part_request_timeout =
-                    match value.get::<i64>().expect("type checked upstream") {
-                        -1 => None,
-                        v => Some(Duration::from_millis(v as u64)),
-                    }
+                    duration_from_millis(value.get::<i64>().expect("type checked upstream"));
             }
             "upload-part-retry-duration" => {
                 settings.upload_part_retry_duration =
-                    match value.get::<i64>().expect("type checked upstream") {
-                        -1 => None,
-                        v => Some(Duration::from_millis(v as u64)),
-                    }
+                    duration_from_millis(value.get::<i64>().expect("type checked upstream"));
+            }
+            "complete-upload-request-timeout" => {
+                settings.complete_upload_request_timeout =
+                    duration_from_millis(value.get::<i64>().expect("type checked upstream"));
+            }
+            "complete-upload-retry-duration" => {
+                settings.complete_upload_retry_duration =
+                    duration_from_millis(value.get::<i64>().expect("type checked upstream"));
             }
             _ => unimplemented!(),
         }
@@ -788,19 +889,19 @@ impl ObjectImpl for S3Sink {
             "secret-access-key" => settings.secret_access_key.to_value(),
             "metadata" => settings.metadata.to_value(),
             "on-error" => settings.multipart_upload_on_error.to_value(),
+            "request-timeout" => duration_to_millis(settings.request_timeout).to_value(),
+            "retry-duration" => duration_to_millis(settings.retry_duration).to_value(),
             "upload-part-request-timeout" => {
-                let timeout: i64 = match settings.upload_part_request_timeout {
-                    None => -1,
-                    Some(v) => v.as_millis() as i64,
-                };
-                timeout.to_value()
+                duration_to_millis(settings.upload_part_request_timeout).to_value()
             }
             "upload-part-retry-duration" => {
-                let timeout: i64 = match settings.upload_part_retry_duration {
-                    None => -1,
-                    Some(v) => v.as_millis() as i64,
-                };
-                timeout.to_value()
+                duration_to_millis(settings.upload_part_retry_duration).to_value()
+            }
+            "complete-upload-request-timeout" => {
+                duration_to_millis(settings.complete_upload_request_timeout).to_value()
+            }
+            "complete-upload-retry-duration" => {
+                duration_to_millis(settings.complete_upload_retry_duration).to_value()
             }
             _ => unimplemented!(),
         }
